---
title: Metric Event
description: A detailed guide on Vector's internal metric data model.
---

import SVG from 'react-inlinesvg';

<SVG src="/img/data-model-metric.svg" />

<!--
     THIS FILE IS AUTOGENERATED!

     To make changes please edit the template located at:

     website/docs/about/data-model/metric.md.erb
-->

## Description

A `metric` event represents a numerical operation to a time series. Operations
offered are heavily inspired by the StatsD and Prometheus models, and determine
the schema of the metric structure within Vector.

When a metric event is sent to a sink the schema will be translated into the
closest equivalent format of the sink protocol.

## Examples

import Tabs from '@theme/Tabs';
import TabItem from '@theme/TabItem';

<Tabs
  defaultValue="counter"
  select={true}
  values={[
    { label: 'Counter', value: 'counter', },
    { label: 'Gauge', value: 'gauge', },
    { label: 'Set', value: 'set', },
    { label: 'Distribution', value: 'distribution', },
    { label: 'Aggregated Histogram', value: 'aggregated_histogram', },
    { label: 'Aggregated Summary', value: 'aggregated_summary', },
  ]
}>
<TabItem value="counter">

```json
{
  "name": "login.count",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "kind": "absolute",
  "tags": {
    "host": "my.host.com"
  },
  "value": {
    "type": "counter",
    "value": 24.2
  }
}
```

</TabItem>
<TabItem value="gauge">

```json
{
  "name": "memory_rss",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "kind": "absolute",
  "tags": {
    "host": "my.host.com"
  },
  "value": {
    "type": "gauge",
    "value": 51200000000.0
  }
}
```

</TabItem>
<TabItem value="set">

```json
{
  "name": "user_names",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "kind": "absolute",
  "tags": {
    "host": "my.host.com"
  },
  "value": {
    "type": "set",
    "values": ["bob", "sam", "ben"]
  }
}
```

</TabItem>
<TabItem value="distribution">

```json
{
  "name": "response_time_ms",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "kind": "absolute",
  "tags": {
    "host": "my.host.com"
  },
  "value": {
    "type": "distribution",
    "values": [2.21, 5.46, 10.22],
    "sample_rates": [5, 2, 5]
  }
}
```

</TabItem>
<TabItem value="aggregated_histogram">

```json
{
  "name": "response_time_ms",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "kind": "absolute",
  "tags": {
    "host": "my.host.com"
  },
  "value": {
    "type": "aggregated_histogram",
    "buckets": [1.0, 2.0, 4.0, 8.0, 16.0, 32.0],
    "counts": [20, 10, 45, 12, 18, 92],
    "count": 197,
    "sum": 975.2
  }
}
```

</TabItem>
<TabItem value="aggregated_summary">

```json
{
  "name": "response_time_ms",
  "timestamp": "2019-11-01T21:15:47+00:00",
  "kind": "absolute",
  "tags": {
    "host": "my.host.com"
  },
  "value": {
    "type": "aggregated_summary",
    "quantiles": [0.1, 0.25, 0.5, 0.9, 0.99, 1.0],
    "values": [2, 3, 5, 8, 9, 10],
    "count": 197,
    "sum": 975.2
  }
}
```

</TabItem>
</Tabs>

## Schema

The metric data model is comprised of 6 types: [`aggregated_histogram`](#aggregated_histogram), [`aggregated_summary`](#aggregated_summary), [`counter`](#counter), [`distribution`](#distribution), [`gauge`](#gauge), [`set`](#set).
You'll notice that certain fields are shared across all types.

import Fields from '@site/src/components/Fields';

import Field from '@site/src/components/Field';

<Fields filters={true}>


<Field
  defaultValue={null}
  enumValues={{"absolute":"The value is an absolute, stand-alone value. It can be used individually.","incremental":"The value is incremental and is used to form a holistic value by merging with other incremental values. Individually it does not tell the whole story."}}
  examples={{"absolute":"The value is an absolute, stand-alone value. It can be used individually.","incremental":"The value is incremental and is used to form a holistic value by merging with other incremental values. Individually it does not tell the whole story."}}
  name={"kind"}
  path={null}
  required={true}
  type={"string"}
  >

### kind

The metric value kind. This determines how the value is merged downstream if metrics are aggregated.



</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={["login.count","response_time"]}
  name={"name"}
  path={null}
  required={true}
  type={"string"}
  >

### name

The metric name.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[{"host":"my.host.com"}]}
  name={"tags"}
  path={null}
  required={true}
  type={"map"}
  >

### tags

Tags that add additional metadata or context to the metric. These are simple key/value pairs in `string` format and cannot be nested.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={["2019-11-01T21:15:47+00:00"]}
  name={"timestamp"}
  path={null}
  required={true}
  type={"timestamp"}
  >

### timestamp

The metric timestamp, representing when the metric was created/ingested within Vector.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"type"}
  path={null}
  required={true}
  type={"struct"}
  >

### type

A metric must be one of 6 types.

<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"aggregated_histogram"}
  path={"type"}
  required={false}
  type={"struct"}
  >

#### aggregated_histogram

Also called a "timer". A [`aggregated_histogram`](#aggregated_histogram) samples observations (usually things like request durations or response sizes) and counts them in configurable buckets. It also provides a sum of all observed values.

<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[[1,2,5,10,25]]}
  name={"buckets"}
  path={"type.aggregated_histogram"}
  required={true}
  type={"[double]"}
  >

##### buckets

The buckets contained within this histogram.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[54]}
  name={"count"}
  path={"type.aggregated_histogram"}
  required={true}
  type={"int"}
  >

##### count

The total number of values contained within the histogram.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[[1,5,25,2,5]]}
  name={"counts"}
  path={"type.aggregated_histogram"}
  required={true}
  type={"[int]"}
  >

##### counts

The number of values contained within each bucket.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[524.0]}
  name={"sum"}
  path={"type.aggregated_histogram"}
  required={true}
  type={"double"}
  >

##### sum

The sum of all values contained within the histogram.


</Field>


</Fields>

</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"aggregated_summary"}
  path={"type"}
  required={false}
  type={"struct"}
  >

#### aggregated_summary

Similar to a histogram, a summary samples observations (usually things like request durations and response sizes). While it also provides a total count of observations and a sum of all observed values, it calculates configurable quantiles over a sliding time window.


<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[54]}
  name={"count"}
  path={"type.aggregated_summary"}
  required={true}
  type={"int"}
  >

##### count

The total number of values contained within the summary.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[[0.1,0.5,0.75,1.0]]}
  name={"quantiles"}
  path={"type.aggregated_summary"}
  required={true}
  type={"[double]"}
  >

##### quantiles

The quantiles contained within the summary, where where 0 ≤ quantile ≤ 1.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[524.0]}
  name={"sum"}
  path={"type.aggregated_summary"}
  required={true}
  type={"double"}
  >

##### sum

The sum of all values contained within the summary.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[[2.1,4.68,23.02,120.1]]}
  name={"values"}
  path={"type.aggregated_summary"}
  required={true}
  type={"[double]"}
  >

##### values

The values contained within the summary that align with the [`quantiles`](#quantiles).


</Field>


</Fields>

</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"counter"}
  path={"type"}
  required={false}
  type={"struct"}
  >

#### counter

A single value that can _only_ be incremented or reset to zero value, it cannot be decremented.

<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[2.6,5.0]}
  name={"value"}
  path={"type.counter"}
  required={true}
  type={"double"}
  >

##### value

The value to increment the counter by. Can only be positive.


</Field>


</Fields>

</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"distribution"}
  path={"type"}
  required={false}
  type={"struct"}
  >

#### distribution

A dsitribution represents a distribution of sampled values.

<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[[12,43,25]]}
  name={"sample_rates"}
  path={"type.distribution"}
  required={true}
  type={"[int]"}
  >

##### sample_rates

The rate at which each individual value was sampled.


</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[[12.0,43.3,25.2]]}
  name={"values"}
  path={"type.distribution"}
  required={true}
  type={"[double]"}
  >

##### values

The list of values contained within the distribution.


</Field>


</Fields>

</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"gauge"}
  path={"type"}
  required={false}
  type={"struct"}
  >

#### gauge

A gauge represents a point-in-time value that can increase and decrease. Vector's internal gauge type represents changes to that value. Gauges should be used to track fluctuations in values, like current memory or CPU usage.

<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[554222.0]}
  name={"value"}
  path={"type.gauge"}
  required={true}
  type={"double"}
  >

##### value

A specific point-in-time value for the gauge.


</Field>


</Fields>

</Field>


<Field
  defaultValue={null}
  enumValues={null}
  examples={null}
  name={"set"}
  path={"type"}
  required={false}
  type={"struct"}
  >

#### set

A set represents a count of unique values, AKA the cardinality.

<Fields filters={false}>


<Field
  defaultValue={null}
  enumValues={null}
  examples={[["unique item 1","unique item 2"]]}
  name={"values"}
  path={"type.set"}
  required={true}
  type={"[string]"}
  >

##### values

The list of unique values.


</Field>


</Fields>

</Field>


</Fields>

</Field>


</Fields>

## Components

import Jump from '@site/src/components/Jump';

<Jump to="/components/?metric=true">View all metric compatible components</Jump>



